import json
import tempfile
from pathlib import Path
from unittest.mock import patch

import pytest
from apistar.exceptions import ErrorResponse
from PIL import Image
from requests.exceptions import ConnectionError

from worker_file_import.image import check_image, create_image, upload_image

SAMPLES = Path(__file__).absolute().parent / "samples"


@patch("worker_file_import.image.tempfile")
def test_check_valid_image(tempfile_mock, tmpdir):
    """
    Do not re-save valid images
    """
    test_file = {
        "id": "fileid",
        "name": "600x600.jpg",
        "local_path": SAMPLES / "600x600.jpg",
        "content_type": "image/jpeg",
    }
    tempfile_mock.mkdtemp.return_value = tmpdir
    result = check_image(test_file)
    # The original file path is returned
    assert result == [test_file["local_path"]]
    # There is nothing at what would have been the new file path
    empty_path = Path(tmpdir) / "fileid"
    assert not empty_path.is_file()
    with Image.open(test_file["local_path"]) as img:
        assert img.format == "JPEG"


def test_check_image_ioerror():
    with pytest.raises(ValueError, match="File 600x600.jpg is not a valid image"):
        check_image(
            {
                "id": "fileid",
                "name": "600x600.jpg",
                "local_path": SAMPLES / "nope.nope",
            }
        )


@patch("worker_file_import.image.tempfile")
def test_check_valid_non_jpeg_image(tempfile_mock, tmpdir):
    test_file = {
        "id": "fileid",
        "name": "600x600.png",
        "local_path": SAMPLES / "600x600.png",
        "content_type": "image/x-png",
    }
    tempfile_mock.mkdtemp.return_value = tmpdir
    result = check_image(test_file)
    assert result == [test_file["local_path"]]
    with Image.open(test_file["local_path"]) as img:
        assert img.format == "PNG"


@patch("worker_file_import.image.tempfile")
def test_check_image_convert_gif(tempfile_mock, tmpdir):
    tempfile_mock.mkdtemp.return_value = tmpdir
    result = check_image(
        {
            "id": "fileid",
            "name": "600x600.gif",
            "local_path": SAMPLES / "600x600.gif",
            "content_type": "image/gif",
        }
    )
    expected_path = Path(tmpdir) / "fileid.jpg"
    assert result == [expected_path]
    assert expected_path.is_file()
    with Image.open(expected_path) as img:
        assert img.format == "JPEG"


@patch("worker_file_import.image.tempfile")
def test_check_image_convert_tiff(tempfile_mock, tmpdir):
    tempfile_mock.mkdtemp.return_value = tmpdir
    result = check_image(
        {
            "id": "fileid",
            "name": "image.tiff",
            "local_path": SAMPLES / "image.tiff",
            "content_type": "image/tiff",
        }
    )
    expected_path = Path(tmpdir) / "fileid.jpg"
    assert result == [expected_path]
    assert expected_path.is_file()
    with Image.open(expected_path) as img:
        assert img.format == "JPEG"


@patch("worker_file_import.image.tempfile")
def test_check_image_exif_orientation(tempfile_mock, tmpdir):
    tempfile_mock.mkdtemp.return_value = tmpdir
    original_path = SAMPLES / "rotated.jpg"

    with Image.open(original_path) as original_image:
        # 0x0112 is the ID of the EXIF Orientation tag. 6 means "rotate 270° clockwise"
        # See https://www.digicamsoft.com/exif22/exif22/html/exif22_22.htm
        assert original_image.getexif().get(0x0112) == 6
        assert original_image.size == (200, 100)

    result = check_image(
        {
            "id": "fileid",
            "name": "rotated.jpg",
            "local_path": original_path,
            "content_type": "image/jpeg",
        }
    )
    expected_path = Path(tmpdir) / "fileid"
    assert result == [expected_path]
    assert expected_path.is_file()
    with Image.open(expected_path) as img:
        assert img.format == "JPEG"
        # Orientation tag has been removed
        assert img.getexif().get(0x0112) is None
        # Size has changed, implying the image was rotated
        assert img.size == (100, 200)


@patch("worker_file_import.image.tempfile")
def test_check_image_orientation_topleft(tempfile_mock, tmpdir):
    """
    Ensure check_image ignores an EXIF Orientation of 1 as it is a no-op
    """
    tempfile_mock.mkdtemp.return_value = tmpdir
    with Image.open(SAMPLES / "rotated.jpg") as img:
        assert img.size == (200, 100)
        exif = img.getexif()
        exif[0x0112] = 1
        with tempfile.NamedTemporaryFile() as f:
            img.save(f, format="JPEG", exif=exif)
            result = check_image(
                {
                    "id": "fileid",
                    "name": "600x600.jpg",
                    "local_path": Path(f.name),
                    "content_type": "image/jpeg",
                }
            )
    # The original file path is returned
    assert result == [Path(f.name)]
    # There is nothing at what would have been the new file path
    expected_path = Path(tmpdir) / "fileid"
    assert not expected_path.is_file()
    with Image.open(SAMPLES / "rotated.jpg") as img:
        assert img.format == "JPEG"
        # Image size has not changed
        assert img.size == (200, 100)


@patch("worker_file_import.image.hash_image")
def test_create_image(hash_image_mock, responses, mock_worker):
    hash_image_mock.return_value = "01e9327d0e5ab7c516a0792c3662bc65"
    mock_worker.api_client.add_response(
        "CreateImage",
        body={"hash": "01e9327d0e5ab7c516a0792c3662bc65"},
        response={"id": "imageid", "s3_put_url": "https://nerv.co.jp/puturl"},
    )
    responses.add(responses.PUT, "https://nerv.co.jp/puturl", status=201)
    mock_worker.api_client.add_response(
        "PartialUpdateImage",
        id="imageid",
        body={"status": "checked"},
        response={"id": "imageid", "status": "checked"},
    )

    img = create_image(mock_worker.api_client, SAMPLES / "200x200.jpg")
    assert img == "imageid"


@patch("worker_file_import.image.hash_image")
def test_create_image_exists(hash_image_mock, mock_worker):
    hash_image_mock.return_value = "01e9327d0e5ab7c516a0792c3662bc65"
    mock_worker.api_client.add_error_response(
        "CreateImage",
        body={"hash": "01e9327d0e5ab7c516a0792c3662bc65"},
        status_code=400,
        content={"id": "imageid"},
    )
    mock_worker.api_client.add_response(
        "RetrieveImage",
        id="imageid",
        response={"id": "imageid", "s3_url": "https://nerv.co.jp/images/something"},
    )
    mock_worker.api_client.add_response(
        "PartialUpdateImage",
        id="imageid",
        body={"status": "checked"},
        response={"id": "imageid", "status": "checked"},
    )

    img = create_image(mock_worker.api_client, SAMPLES / "200x200.jpg")
    assert img == "imageid"


@patch("worker_file_import.image.hash_image")
def test_create_image_error(hash_image_mock, mock_worker):
    """
    Test that create_image only handles HTTP 400 errors
    """
    hash_image_mock.return_value = "01e9327d0e5ab7c516a0792c3662bc65"
    mock_worker.api_client.add_error_response(
        "CreateImage",
        body={"hash": "01e9327d0e5ab7c516a0792c3662bc65"},
        status_code=500,
        content={"aaaaaaaaaaaaaaaaaaa"},
    )

    with pytest.raises(ErrorResponse):
        create_image(mock_worker.api_client, SAMPLES / "200x200.jpg")


@patch(
    "worker_file_import.image.upload_image.retry.wait.wait_fixed",
    new=0,
)
def test_upload_retry(responses):
    responses.add(responses.PUT, "https://nerv.co.jp/puturl", body=ConnectionError())
    responses.add(responses.PUT, "https://nerv.co.jp/puturl", status=500)
    responses.add(responses.PUT, "https://nerv.co.jp/puturl", status=201)

    upload_image(SAMPLES / "200x200.jpg", "https://nerv.co.jp/puturl")

    assert responses.assert_call_count("https://nerv.co.jp/puturl", 3)


@patch(
    "worker_file_import.image.upload_image.retry.wait.wait_fixed",
    new=0,
)
def test_upload_retry_exhausted(responses):
    responses.add(responses.PUT, "https://nerv.co.jp/puturl", body=ConnectionError())

    with pytest.raises(ConnectionError):
        upload_image(SAMPLES / "200x200.jpg", "https://nerv.co.jp/puturl")

    assert responses.assert_call_count("https://nerv.co.jp/puturl", 3)


@patch("worker_file_import.image.hash_image")
def test_run_image(hash_image_mock, responses, mock_worker):
    hash_image_mock.return_value = "01e9327d0e5ab7c516a0792c3662bc65"
    mock_worker.process_information = {
        "id": "processid",
        "mode": "files",
        "corpus": "corpusid",
        "element": {
            "id": "folderid",
            "name": "Untitled",
            "type": "folder",
            "corpus": {"id": "corpusid"},
        },
        "folder_type": "folder",
        "element_type": "page",
        "files": ["file1"],
    }
    # DataFile info
    mock_worker.api_client.add_response(
        "RetrieveDataFile",
        id="file1",
        response={
            "id": "file1",
            "name": "600x600.jpg",
            "s3_url": "https://nerv.co.jp/600x600.jpg",
            "content_type": "image/jpeg",
        },
    )
    # DataFile on S3
    responses.add(
        responses.GET,
        "https://nerv.co.jp/600x600.jpg",
        body=(SAMPLES / "600x600.jpg").open("rb"),
    )
    # Image creation
    mock_worker.api_client.add_response(
        "CreateImage",
        body={"hash": "01e9327d0e5ab7c516a0792c3662bc65"},
        response={
            "id": "aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa",
            "s3_put_url": "https://nerv.co.jp/put",
            "s3_url": "https://nerv.co.jp/600x600.jpg",
        },
    )
    responses.add(responses.PUT, "https://nerv.co.jp/put", status=201)
    mock_worker.api_client.add_response(
        "PartialUpdateImage",
        id="aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa",
        body={"status": "checked"},
        response={"id": "aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa", "status": "checked"},
    )
    # Page creation
    mock_worker.api_client.add_response(
        "CreateElement",
        body={
            "corpus": "corpusid",
            "type": "page",
            "name": "600x600.jpg",
            "polygon": None,
            "confidence": None,
            "image": "aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa",
            "worker_run_id": "bbbbbbbb-aaaa-aaaa-aaaa-aaaaaaaaaaaa",
            "parent": "folderid",
        },
        response={"id": "pageid", "type": "page", "name": "600x600.jpg"},
    )
    # Process update
    mock_worker.api_client.add_response(
        "PartialUpdateProcess",
        id="processid",
        body={"element_id": "folderid"},
        response={},
    )
    # DataFile deletion
    mock_worker.api_client.add_response("DestroyDataFile", id="file1", response={})

    mock_worker.run()

    with Path.open(mock_worker.work_dir / "elements.json") as f:
        elements = json.load(f)
    assert elements == [
        {"id": "folderid", "type": "folder"},
        {"id": "pageid", "type": "page", "name": "600x600.jpg"},
    ]


@patch("worker_file_import.image.hash_image")
def test_run_top_level(hash_image_mock, responses, mock_worker):
    hash_image_mock.return_value = "01e9327d0e5ab7c516a0792c3662bc65"
    mock_worker.process_information = {
        "id": "processid",
        "corpus": "corpusid",
        "mode": "files",
        "element": None,
        "files": ["file1"],
        "folder_type": "folder",
        "element_type": "page",
    }
    # DataFile info
    mock_worker.api_client.add_response(
        "RetrieveDataFile",
        id="file1",
        response={
            "id": "file1",
            "name": "600x600.jpg",
            "s3_url": "https://nerv.co.jp/600x600.jpg",
            "content_type": "image/jpeg",
        },
    )
    # DataFile on S3
    responses.add(
        responses.GET,
        "https://nerv.co.jp/600x600.jpg",
        body=(SAMPLES / "600x600.jpg").open("rb"),
    )
    # Image creation
    mock_worker.api_client.add_response(
        "CreateImage",
        body={"hash": "01e9327d0e5ab7c516a0792c3662bc65"},
        response={
            "id": "aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa",
            "s3_put_url": "https://nerv.co.jp/put",
            "s3_url": "https://nerv.co.jp/600x600.jpg",
        },
    )
    responses.add(responses.PUT, "https://nerv.co.jp/put", status=201)
    mock_worker.api_client.add_response(
        "PartialUpdateImage",
        id="aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa",
        body={"status": "checked"},
        response={"id": "aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa", "status": "checked"},
    )
    # Folder and page creation
    mock_worker.api_client.add_response(
        "CreateElement",
        body={
            "corpus": "corpusid",
            "type": "folder",
            "name": "600x600.jpg",
            "worker_run_id": "bbbbbbbb-aaaa-aaaa-aaaa-aaaaaaaaaaaa",
        },
        response={
            "id": "folderid",
            "type": "folder",
            "name": "600x600.jpg",
            "corpus": {"id": "corpusid"},
        },
    )
    mock_worker.api_client.add_response(
        "CreateElement",
        body={
            "corpus": "corpusid",
            "type": "page",
            "name": "600x600.jpg",
            "polygon": None,
            "confidence": None,
            "image": "aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa",
            "worker_run_id": "bbbbbbbb-aaaa-aaaa-aaaa-aaaaaaaaaaaa",
            "parent": "folderid",
        },
        response={"id": "pageid", "type": "page", "name": "600x600.jpg"},
    )
    # Process update
    mock_worker.api_client.add_response(
        "PartialUpdateProcess",
        id="processid",
        body={"element_id": "folderid"},
        response={},
    )
    # DataFile deletion
    mock_worker.api_client.add_response("DestroyDataFile", id="file1", response={})

    mock_worker.run()

    with (mock_worker.work_dir / "elements.json").open() as f:
        elements = json.load(f)
    assert elements == [
        {"id": "folderid", "type": "folder"},
        {"id": "pageid", "type": "page", "name": "600x600.jpg"},
    ]


@patch("worker_file_import.image.hash_image")
def test_run_sorted(hash_image_mock, responses, mock_worker):
    """
    Ensure that files are sorted by name using natsort
    """
    hash_image_mock.return_value = "01e9327d0e5ab7c516a0792c3662bc65"

    numbers = [1, 2, 10, 100, 234]
    assert sorted(numbers) == [1, 2, 10, 100, 234]
    assert sorted(map(str, numbers)) == ["1", "10", "100", "2", "234"]

    mock_worker.process_information = {
        "id": "processid",
        "corpus": "corpusid",
        "mode": "files",
        "element": {
            "id": "folderid",
            "name": "Untitled",
            "type": "folder",
            "corpus": {"id": "corpusid"},
        },
        "files": list(map(str, numbers)),
        "folder_type": "folder",
        "element_type": "page",
    }

    # DataFile info and file on S3
    for i in numbers:
        mock_worker.api_client.add_response(
            "RetrieveDataFile",
            id=str(i),
            response={
                "id": str(i),
                "name": f"file_{i}.jpg",
                "s3_url": f"https://nerv.co.jp/file_{i}.jpg",
                "content_type": "image/jpeg",
            },
        )
        responses.add(
            responses.GET,
            f"https://nerv.co.jp/file_{i}.jpg",
            body=(SAMPLES / "600x600.jpg").open("rb"),
        )

    # Image creation
    for i in numbers:
        mock_worker.api_client.add_response(
            "CreateImage",
            body={"hash": "01e9327d0e5ab7c516a0792c3662bc65"},
            response={
                "id": "aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa",
                "s3_put_url": "https://nerv.co.jp/put",
                "s3_url": f"https://nerv.co.jp/file_{i}.jpg",
            },
        )
        responses.add(responses.PUT, "https://nerv.co.jp/put", status=201)
        mock_worker.api_client.add_response(
            "PartialUpdateImage",
            id="aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa",
            body={"status": "checked"},
            response={
                "id": "aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa",
                "status": "checked",
            },
        )

    # Page creation
    for i in numbers:
        mock_worker.api_client.add_response(
            "CreateElement",
            body={
                "corpus": "corpusid",
                "type": "page",
                "name": f"file_{i}.jpg",
                "polygon": None,
                "confidence": None,
                "image": "aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa",
                "worker_run_id": "bbbbbbbb-aaaa-aaaa-aaaa-aaaaaaaaaaaa",
                "parent": "folderid",
            },
            response={"id": "pageid", "type": "page", "name": f"file_{i}.jpg"},
        )

    # Process element update
    mock_worker.api_client.add_response(
        "PartialUpdateProcess",
        id="processid",
        body={"element_id": "folderid"},
        response={},
    )

    # DataFile deletion
    for i in numbers:
        mock_worker.api_client.add_response("DestroyDataFile", id=str(i), response={})

    mock_worker.run()

    with (mock_worker.work_dir / "elements.json").open() as f:
        elements = json.load(f)
    assert elements == [
        {"id": "folderid", "type": "folder"},
        {"id": "pageid", "type": "page", "name": "file_1.jpg"},
        {"id": "pageid", "type": "page", "name": "file_2.jpg"},
        {"id": "pageid", "type": "page", "name": "file_10.jpg"},
        {"id": "pageid", "type": "page", "name": "file_100.jpg"},
        {"id": "pageid", "type": "page", "name": "file_234.jpg"},
    ]
