import json
from datetime import datetime
from pathlib import Path
from unittest.mock import call, patch

from arkindex_worker.worker import Element

SAMPLES = Path(__file__).absolute().parent / "samples"


@patch("worker_file_import.worker.check_image")
def test_get_images(check_image_mock, mock_worker):
    check_image_mock.return_value = [Path("/something.jpg")]
    test_datafile = {
        "id": "file1",
        "name": "file1.jpg",
        "local_path": "some/path/file1.jpg",
        "content_type": "image/jpeg",
    }
    mock_worker.process_information = {
        "id": "processid",
        "mode": "files",
        "element_type": "page",
        "element": {"id": "folderid"},
    }
    mock_worker.get_images(test_datafile)

    assert check_image_mock.call_count == 1
    assert check_image_mock.call_args == call(test_datafile)


@patch("worker_file_import.worker.extract_pdf_images")
def test_get_images_pdf(extract_pdf_images_mock, mock_worker):
    extract_pdf_images_mock.return_value = [Path("/something.jpg")]
    test_datafile = {
        "id": "file1",
        "name": "file1.pdf",
        "local_path": "some/path/file1.pdf",
        "content_type": "application/pdf",
    }

    mock_worker.process_information = {
        "id": "processid",
        "mode": "files",
        "element_type": "page",
        "element": {"id": "folderid"},
    }
    mock_worker.get_images(test_datafile)

    assert extract_pdf_images_mock.call_count == 1
    assert extract_pdf_images_mock.call_args == call(test_datafile)


@patch("worker_file_import.worker.check_image")
def test_get_images_error(check_image_mock, mock_worker):
    check_image_mock.side_effect = FileNotFoundError
    mock_worker.process_information = {
        "id": "processid",
        "mode": "files",
        "element_type": "page",
        "element": {"id": "folderid"},
    }
    datafile = {"id": "file1", "name": "file.jpg", "content_type": "image/jpeg"}

    assert mock_worker.get_images(datafile) == []

    assert check_image_mock.call_count == 1
    assert check_image_mock.call_args == call(datafile)


@patch("worker_file_import.worker.datetime")
def test_create_top_level_folder(datetime_mock, mock_worker):
    mock_worker.process_information = {
        "id": "processid",
        "mode": "files",
        "corpus": "corpusid",
        "element": None,
        "folder_type": "folder",
        "element_type": "page",
        "files": ["file1", "file2"],
    }
    datetime_mock.now.return_value = datetime(2020, 2, 2, 20, 2)
    mock_worker.api_client.add_response(
        "CreateElement",
        body={
            "corpus": "corpusid",
            "type": "folder",
            "name": "Import images 2020-02-02 20:02",
            "worker_run_id": "bbbbbbbb-aaaa-aaaa-aaaa-aaaaaaaaaaaa",
        },
        response={
            "id": "folderid",
            "type": "folder",
            "name": "Import images 2020-02-02 20:02",
        },
    )

    mock_worker.create_top_level_folder(
        [
            {"id": "file1", "name": "file.jpg", "content_type": "image/jpeg"},
            {"id": "file2", "name": "file2.jpg", "content_type": "image/jpeg"},
        ]
    )
    assert mock_worker.process_information["element"] == {
        "id": "folderid",
        "type": "folder",
        "name": "Import images 2020-02-02 20:02",
    }


def test_create_top_level_folder_datafile_name(mock_worker):
    mock_worker.process_information = {
        "id": "processid",
        "mode": "files",
        "corpus": "corpusid",
        "element": None,
        "folder_type": "folder",
        "element_type": "page",
        "files": ["file1"],
    }
    mock_worker.api_client.add_response(
        "CreateElement",
        body={
            "corpus": "corpusid",
            "type": "folder",
            "name": "file.jpg",
            "worker_run_id": "bbbbbbbb-aaaa-aaaa-aaaa-aaaaaaaaaaaa",
        },
        response={"id": "folderid", "type": "folder", "name": "file.jpg"},
    )

    mock_worker.create_top_level_folder(
        [{"id": "file1", "name": "file.jpg", "content_type": "image/jpeg"}]
    )

    assert mock_worker.process_information["element"] == {
        "id": "folderid",
        "type": "folder",
        "name": "file.jpg",
    }


def test_create_top_level_folder_ignore_element(mock_worker):
    mock_worker.process_information = {
        "id": "processid",
        "mode": "files",
        "corpus": "corpusid",
        "element": {"id": "folderid", "name": "My folder"},
        "folder_type": "folder",
        "element_type": "page",
        "files": ["file1"],
    }
    # No mocked API response as there should be no API call
    mock_worker.create_top_level_folder([{"id": "file1", "name": "file.jpg"}])


def test_create_top_level_folder_ignore_pdf(mock_worker):
    mock_worker.process_information = {
        "id": "processid",
        "mode": "files",
        "corpus": "corpusid",
        "element": None,
        "folder_type": "folder",
        "element_type": "page",
        "files": ["file1"],
    }
    # No mocked API response as there should be no API call
    mock_worker.create_top_level_folder(
        [{"id": "file1", "name": "file.pdf", "content_type": "application/pdf"}]
    )
    assert mock_worker.process_information["element"] is None


@patch("worker_file_import.image.hash_image")
def test_save_elements(hash_image_mock, responses, mock_worker):
    hash_image_mock.return_value = "01e9327d0e5ab7c516a0792c3662bc65"
    mock_worker.process_information = {
        "id": "processid",
        "corpus": "corpusid",
        "element": {"id": "folderid", "type": "folder", "corpus": {"id": "corpusid"}},
        "folder_type": "folder",
        "element_type": "page",
    }

    mock_worker.api_client.add_response(
        "CreateImage",
        body={"hash": "01e9327d0e5ab7c516a0792c3662bc65"},
        response={
            "id": "aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa",
            "s3_put_url": "https://nerv.co.jp/puturl",
            "s3_url": "http://s3/200x200.jpg",
        },
    )
    responses.add(responses.PUT, "https://nerv.co.jp/puturl", status=201)
    mock_worker.api_client.add_response(
        "PartialUpdateImage",
        id="aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa",
        body={"status": "checked"},
        response={"id": "aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa", "status": "checked"},
    )
    mock_worker.api_client.add_response(
        "CreateElement",
        body={
            "corpus": "corpusid",
            "type": "page",
            "name": "200x200.jpg",
            "polygon": None,
            "confidence": None,
            "image": "aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa",
            "worker_run_id": "bbbbbbbb-aaaa-aaaa-aaaa-aaaaaaaaaaaa",
            "parent": "folderid",
        },
        response={"id": "pageid", "type": "page", "name": "200x200.jpg"},
    )

    datafile = {
        "id": "file1",
        "name": "200x200.jpg",
        "s3_url": "http://s3/200x200.jpg",
    }
    pages = mock_worker.save_elements(
        datafile,
        [SAMPLES / "200x200.jpg"],
        parent_element=Element(
            {"id": "folderid", "type": "folder", "corpus": {"id": "corpusid"}}
        ),
    )

    assert pages == [{"id": "pageid", "type": "page", "name": "200x200.jpg"}]


@patch("worker_file_import.image.hash_image")
def test_save_elements_top_level(hash_image_mock, responses, mock_worker):
    """
    Assert that save_elements creates folders even when a datafile has a single image
    if the process runs on a corpus, without a parent element
    """
    hash_image_mock.return_value = "01e9327d0e5ab7c516a0792c3662bc65"
    mock_worker.process_information = {
        "id": "processid",
        "corpus": "corpusid",
        "element": None,
        "folder_type": "folder",
        "element_type": "page",
    }

    mock_worker.api_client.add_response(
        "CreateImage",
        body={"hash": "01e9327d0e5ab7c516a0792c3662bc65"},
        response={
            "id": "aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa",
            "s3_put_url": "https://nerv.co.jp/puturl",
            "s3_url": "http://s3/200x200.jpg",
        },
    )
    responses.add(responses.PUT, "https://nerv.co.jp/puturl", status=201)
    mock_worker.api_client.add_response(
        "PartialUpdateImage",
        id="aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa",
        body={"status": "checked"},
        response={"id": "aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa", "status": "checked"},
    )
    mock_worker.api_client.add_response(
        "CreateElement",
        body={
            "corpus": "corpusid",
            "type": "folder",
            "name": "200x200.jpg",
            "worker_run_id": "bbbbbbbb-aaaa-aaaa-aaaa-aaaaaaaaaaaa",
            "parent": None,
        },
        response={
            "id": "folderid",
            "type": "folder",
            "name": "200x200.jpg",
            "corpus": {"id": "corpusid"},
        },
    )
    mock_worker.api_client.add_response(
        "CreateElement",
        body={
            "corpus": "corpusid",
            "type": "page",
            "name": "1",
            "polygon": None,
            "confidence": None,
            "image": "aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa",
            "worker_run_id": "bbbbbbbb-aaaa-aaaa-aaaa-aaaaaaaaaaaa",
            "parent": "folderid",
        },
        response={"id": "pageid", "type": "page", "name": "1"},
    )

    datafile = {
        "id": "file1",
        "name": "200x200.jpg",
        "s3_url": "http://s3/200x200.jpg",
    }
    elements = mock_worker.save_elements(
        datafile, [SAMPLES / "200x200.jpg"], parent_element=None
    )

    assert elements == [
        {
            "id": "folderid",
            "type": "folder",
            "name": "200x200.jpg",
            "corpus": {"id": "corpusid"},
        },
        {"id": "pageid", "type": "page", "name": "1"},
    ]


@patch("worker_file_import.image.hash_image")
def test_run_pdf_and_image(hash_image_mock, mock_worker, responses):
    hash_image_mock.return_value = "01e9327d0e5ab7c516a0792c3662bc65"
    # Process info
    mock_worker.process_information = {
        "id": "processid",
        "corpus": "corpusid",
        "mode": "files",
        "element": {
            "id": "folderid",
            "name": "Untitled",
            "type": "book",
            "corpus": {"id": "corpusid"},
        },
        "files": ["file1", "file2"],
        "element_type": "page",
        "folder_type": "book",
    }

    # DataFiles info
    mock_worker.api_client.add_response(
        "RetrieveDataFile",
        id="file1",
        response={
            "id": "file1",
            "name": "file.pdf",
            "content_type": "application/pdf",
            "s3_url": "http://s3/file.pdf",
        },
    )
    mock_worker.api_client.add_response(
        "RetrieveDataFile",
        id="file2",
        response={
            "id": "file2",
            "name": "600x600.jpg",
            "s3_url": "https://nerv.co.jp/600x600.jpg",
            "content_type": "image/jpeg",
        },
    )

    # DataFiles on S3
    responses.add(
        responses.GET,
        "https://nerv.co.jp/600x600.jpg",
        body=(SAMPLES / "600x600.jpg").open("rb"),
    )
    responses.add(
        responses.GET, "http://s3/file.pdf", body=(SAMPLES / "file.pdf").open("rb")
    )

    # Images creation
    mock_worker.api_client.add_response(
        "CreateImage",
        body={"hash": "01e9327d0e5ab7c516a0792c3662bc65"},
        response={
            "id": "aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa",
            "s3_put_url": "https://nerv.co.jp/put",
            "s3_url": "https://nerv.co.jp/600x600.jpg",
        },
    )
    responses.add(responses.PUT, "https://nerv.co.jp/put", status=201)
    mock_worker.api_client.add_response(
        "PartialUpdateImage",
        id="aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa",
        body={"status": "checked"},
        response={"id": "aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa", "status": "checked"},
    )

    mock_worker.api_client.add_response(
        "CreateImage",
        body={"hash": "01e9327d0e5ab7c516a0792c3662bc65"},
        response={
            "id": "bbbbbbbb-aaaa-3333-aaaa-aaaaaaaaaaaa",
            "s3_put_url": "http://s3/put",
            "s3_url": "http://s3/image1.jpg",
        },
    )
    mock_worker.api_client.add_response(
        "CreateImage",
        body={"hash": "01e9327d0e5ab7c516a0792c3662bc65"},
        response={
            "id": "bbbbbbbb-aaaa-4444-aaaa-aaaaaaaaaaaa",
            "s3_put_url": "http://s3/put",
            "s3_url": "http://s3/image2.jpg",
        },
    )
    responses.add(responses.PUT, "http://s3/put", status=201)
    responses.add(responses.PUT, "http://s3/put", status=201)
    mock_worker.api_client.add_response(
        "PartialUpdateImage",
        id="bbbbbbbb-aaaa-4444-aaaa-aaaaaaaaaaaa",
        body={"status": "checked"},
        response={"id": "bbbbbbbb-aaaa-4444-aaaa-aaaaaaaaaaaa", "status": "checked"},
    )
    mock_worker.api_client.add_response(
        "PartialUpdateImage",
        id="bbbbbbbb-aaaa-3333-aaaa-aaaaaaaaaaaa",
        body={"status": "checked"},
        response={"id": "bbbbbbbb-aaaa-3333-aaaa-aaaaaaaaaaaa", "status": "checked"},
    )

    # Elements creation
    mock_worker.api_client.add_response(
        "CreateElement",
        body={
            "corpus": "corpusid",
            "type": "page",
            "name": "600x600.jpg",
            "polygon": None,
            "confidence": None,
            "image": "aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa",
            "worker_run_id": mock_worker.worker_run_id,
            "parent": "folderid",
        },
        response={"id": "pageid", "type": "page", "name": "600x600.jpg"},
    )
    mock_worker.api_client.add_response(
        "CreateElement",
        body={
            "parent": "folderid",
            "corpus": "corpusid",
            "type": "book",
            "name": "file.pdf",
            "worker_run_id": mock_worker.worker_run_id,
        },
        response={
            "id": "box1",
            "type": "book",
            "corpus": {"id": "corpusid"},
        },
    )
    mock_worker.api_client.add_response(
        "CreateElement",
        body={
            "type": "page",
            "name": "1",
            "image": "bbbbbbbb-aaaa-3333-aaaa-aaaaaaaaaaaa",
            "corpus": "corpusid",
            "polygon": None,
            "parent": "box1",
            "worker_run_id": mock_worker.worker_run_id,
            "confidence": None,
        },
        response={
            "id": "page1",
            "type": "page",
            "zone": {
                "image": {
                    "width": 1200,
                    "height": 3000,
                }
            },
        },
    )
    mock_worker.api_client.add_response(
        "CreateElement",
        body={
            "type": "page",
            "name": "2",
            "image": "bbbbbbbb-aaaa-4444-aaaa-aaaaaaaaaaaa",
            "corpus": "corpusid",
            "polygon": None,
            "parent": "box1",
            "worker_run_id": mock_worker.worker_run_id,
            "confidence": None,
        },
        response={
            "id": "page2",
            "type": "page",
            "zone": {
                "image": {
                    "width": 1200,
                    "height": 3000,
                }
            },
        },
    )

    # Process update
    mock_worker.api_client.add_response(
        "PartialUpdateProcess",
        id="processid",
        body={"element_id": "folderid"},
        response={},
    )

    # DataFile deletion
    mock_worker.api_client.add_response("DestroyDataFile", id="file1", response={})
    mock_worker.api_client.add_response("DestroyDataFile", id="file2", response={})

    # PDF transcriptions creation
    mock_worker.api_client.add_response(
        "CreateElementTranscriptions",
        id="page1",
        body={
            "element_type": "text_line",
            "worker_run_id": mock_worker.worker_run_id,
            "transcriptions": [
                {
                    "text": " A Simple PDF File \n",
                    "polygon": [[112, 285], [574, 285], [574, 183], [112, 183]],
                    "confidence": 1,
                    "orientation": "horizontal-lr",
                },
                {
                    "text": " This is a small demonstration .pdf file - \n",
                    "polygon": [[136, 399], [481, 399], [481, 362], [136, 362]],
                    "confidence": 1,
                    "orientation": "horizontal-lr",
                },
                {
                    "text": " just for use in the Virtual Mechanics tutorials. More text. And more \n",
                    "polygon": [[136, 490], [719, 490], [719, 452], [136, 452]],
                    "confidence": 1,
                    "orientation": "horizontal-lr",
                },
                {
                    "text": " text. And more text. And more text. And more text. \n",
                    "polygon": [[136, 535], [583, 535], [583, 497], [136, 497]],
                    "confidence": 1,
                    "orientation": "horizontal-lr",
                },
                {
                    "text": " And more text. And more text. And more text. And more text. And more \n",
                    "polygon": [[136, 626], [764, 626], [764, 588], [136, 588]],
                    "confidence": 1,
                    "orientation": "horizontal-lr",
                },
                {
                    "text": " text. And more text. Boring, zzzzz. And more text. And more text. And \n",
                    "polygon": [[136, 671], [750, 671], [750, 633], [136, 633]],
                    "confidence": 1,
                    "orientation": "horizontal-lr",
                },
                {
                    "text": " more text. And more text. And more text. And more text. And more text. \n",
                    "polygon": [[136, 716], [766, 716], [766, 679], [136, 679]],
                    "confidence": 1,
                    "orientation": "horizontal-lr",
                },
                {
                    "text": " And more text. And more text. \n",
                    "polygon": [[136, 762], [407, 762], [407, 724], [136, 724]],
                    "confidence": 1,
                    "orientation": "horizontal-lr",
                },
                {
                    "text": " And more text. And more text. And more text. And more text. And more \n",
                    "polygon": [[136, 852], [764, 852], [764, 814], [136, 814]],
                    "confidence": 1,
                    "orientation": "horizontal-lr",
                },
                {
                    "text": " text. And more text. And more text. Even more. Continued on page 2 ...\n",
                    "polygon": [[136, 897], [759, 897], [759, 860], [136, 860]],
                    "confidence": 1,
                    "orientation": "horizontal-lr",
                },
            ],
            "return_elements": True,
        },
        response={},
    )
    mock_worker.api_client.add_response(
        "CreateElementTranscriptions",
        id="page2",
        body={
            "element_type": "text_line",
            "worker_run_id": mock_worker.worker_run_id,
            "transcriptions": [
                {
                    "text": " Simple PDF File 2 \n",
                    "polygon": [[112, 285], [568, 285], [568, 183], [112, 183]],
                    "confidence": 1,
                    "orientation": "horizontal-lr",
                },
                {
                    "text": " ...continued from page 1. Yet more text. And more text. And more text. \n",
                    "polygon": [[136, 399], [757, 399], [757, 362], [136, 362]],
                    "confidence": 1,
                    "orientation": "horizontal-lr",
                },
                {
                    "text": " And more text. And more text. And more text. And more text. And more \n",
                    "polygon": [[136, 445], [764, 445], [764, 407], [136, 407]],
                    "confidence": 1,
                    "orientation": "horizontal-lr",
                },
                {
                    "text": " text. Oh, how boring typing this stuff. But not as boring as watching \n",
                    "polygon": [[136, 490], [727, 490], [727, 452], [136, 452]],
                    "confidence": 1,
                    "orientation": "horizontal-lr",
                },
                {
                    "text": " paint dry. And more text. And more text. And more text. And more text. \n",
                    "polygon": [[136, 535], [759, 535], [759, 497], [136, 497]],
                    "confidence": 1,
                    "orientation": "horizontal-lr",
                },
                {
                    "text": " Boring.  More, a little more text. The end, and just as well. \n",
                    "polygon": [[136, 581], [647, 581], [647, 543], [136, 543]],
                    "confidence": 1,
                    "orientation": "horizontal-lr",
                },
            ],
            "return_elements": True,
        },
        response={},
    )

    mock_worker.run()

    with (mock_worker.work_dir / "elements.json").open() as f:
        elements = json.load(f)
    assert elements == [
        {"id": "folderid", "type": "book"},
        {"id": "pageid", "type": "page", "name": "600x600.jpg"},
        {"id": "box1", "type": "book", "corpus": {"id": "corpusid"}},
        {
            "id": "page1",
            "type": "page",
            "zone": {
                "image": {
                    "width": 1200,
                    "height": 3000,
                }
            },
        },
        {
            "id": "page2",
            "type": "page",
            "zone": {
                "image": {
                    "width": 1200,
                    "height": 3000,
                }
            },
        },
    ]
