import json
from pathlib import Path
from unittest.mock import MagicMock, call, patch

import pytest

from arkindex_worker.models import Element
from arkindex_worker.worker.transcription import TextOrientation
from worker_file_import.pdf import (
    build_transcription,
    extract_pdf_images,
    extract_pdf_text,
)

SAMPLES = Path(__file__).absolute().parent / "samples"

DATAFILE = {
    "id": "datafileid",
    "name": "file.pdf",
    "content_type": "application/pdf",
    "local_path": SAMPLES / "file.pdf",
}
ARABIC_DATAFILE = {
    "id": "arabicid",
    "name": "arabic_sample.pdf",
    "content_type": "application/pdf",
    "local_path": SAMPLES / "arabic_sample.pdf",
}


@patch("worker_file_import.pdf.convert_from_path")
def test_extract_pdf_images_not_pdf(conv_mock):
    bad_datafile = DATAFILE.copy()
    bad_datafile["content_type"] = "text/plain"

    with pytest.raises(ValueError, match="File file.pdf is not a PDF"):
        extract_pdf_images(bad_datafile)

    assert not conv_mock.called


@patch("worker_file_import.pdf.shutil.which")
def test_extract_pdf_images_missing_executable(find_exec_mock):
    find_exec_mock.return_value = False

    with pytest.raises(AssertionError, match="Missing pdftoppm in PATH"):
        extract_pdf_images(DATAFILE)

    assert find_exec_mock.call_count == 1
    assert find_exec_mock.call_args == call("pdftoppm")


def test_extract_pdf_images(tmpdir):
    result = extract_pdf_images(DATAFILE, tmpdir)
    assert result == [Path(tmpdir / "pdf-0001-1.jpg"), Path(tmpdir / "pdf-0001-2.jpg")]


def test_extract_pdf_text():
    ark_pages = [
        Element(
            {
                "id": "page1",
                "type": "page",
                "zone": {
                    "image": {
                        "width": 1000,
                        "height": 2000,
                    }
                },
            }
        ),
        Element(
            {
                "id": "page2",
                "type": "page",
                "zone": {
                    "image": {
                        "width": 800,
                        "height": 1900,
                    }
                },
            }
        ),
    ]

    result = extract_pdf_text(DATAFILE["local_path"], ark_pages)

    assert result[0].page_element.id == "page1"
    assert len(result[0].transcriptions) == 10
    assert result[1].page_element.id == "page2"
    assert len(result[1].transcriptions) == 6

    check = SAMPLES / "file_text.json"
    expected = json.loads(check.read_text())
    for item in expected:
        for tr in item["transcriptions"]:
            tr["orientation"] = TextOrientation(tr["orientation"])
    assert expected == [item._asdict() for item in result]


def test_extract_pdf_text_orientation():
    ark_page = [
        Element(
            {
                "id": "page1",
                "type": "page",
                "zone": {
                    "image": {
                        "width": 1000,
                        "height": 2000,
                    }
                },
            }
        )
    ]

    result = extract_pdf_text(ARABIC_DATAFILE["local_path"], ark_page)
    assert result[0].page_element.id == "page1"
    assert len(result[0].transcriptions) == 170

    check = SAMPLES / "arabic_sample.json"
    expected = json.loads(check.read_text())
    for item in expected:
        for tr in item["transcriptions"]:
            tr["orientation"] = TextOrientation(tr["orientation"])
    assert expected == [item._asdict() for item in result]


def test_build_transcription_wrong_polygon():
    pdf_element = MagicMock()
    pdf_element.x0 = 0
    pdf_element.x1 = 1
    pdf_element.y0 = 1
    pdf_element.y1 = 1
    pdf_element.get_text.return_value = "hello"

    pdf_page = MagicMock()
    pdf_page.width = 500
    pdf_page.height = 1000

    ark_page = Element(
        {
            "id": "page1",
            "type": "page",
            "zone": {
                "image": {
                    "width": 1000,
                    "height": 2000,
                }
            },
        }
    )

    # Checking that we don't have None because the text could be empty
    assert pdf_element.get_text().strip() is not None
    result = build_transcription(pdf_element, pdf_page, ark_page)
    assert result is None


def test_build_transcription_polygon_out_of_bounds():
    pdf_element = MagicMock()
    pdf_element.x0 = -50
    pdf_element.x1 = 600
    pdf_element.y0 = -9999
    pdf_element.y1 = 32135
    pdf_element.get_text.return_value = "hello"

    pdf_page = MagicMock()
    pdf_page.width = 500
    pdf_page.height = 1000

    ark_page = Element(
        {
            "id": "page1",
            "type": "page",
            "zone": {
                "image": {
                    "width": 1000,
                    "height": 2000,
                }
            },
        }
    )

    # Checking that we don't have None because the text could be empty
    assert pdf_element.get_text().strip() is not None
    result = build_transcription(pdf_element, pdf_page, ark_page)
    assert result == {
        "text": "hello",
        "polygon": [
            [0, 2000],
            [1000, 2000],
            [1000, 0],
            [0, 0],
        ],
        "confidence": 1,
        "orientation": TextOrientation.HorizontalLeftToRight,
    }


@patch("worker_file_import.image.hash_image")
def test_run_pdf(hash_image_mock, mock_worker, responses):
    hash_image_mock.return_value = "01e9327d0e5ab7c516a0792c3662bc65"
    mock_worker.process_information = {
        "id": "processid",
        "corpus": "corpusid",
        "mode": "files",
        "element": {"id": "folderid", "name": "Untitled", "type": "book"},
        "files": ["file1", "file2"],
        "element_type": "page",
        "folder_type": "book",
    }

    # DataFile info
    mock_worker.api_client.add_response(
        "RetrieveDataFile",
        id="file1",
        response={
            "id": "file1",
            "name": "file.pdf",
            "content_type": "application/pdf",
            "s3_url": "http://s3/file.pdf",
        },
    )
    mock_worker.api_client.add_response(
        "RetrieveDataFile",
        id="file2",
        response={
            "id": "file2",
            "name": "file2.pdf",
            "content_type": "application/pdf",
            "s3_url": "http://s3/file2.pdf",
        },
    )

    # DataFile on S3
    responses.add(
        responses.GET, "http://s3/file.pdf", body=(SAMPLES / "file.pdf").open("rb")
    )
    responses.add(
        responses.GET, "http://s3/file2.pdf", body=(SAMPLES / "file2.pdf").open("rb")
    )

    # Image creation
    mock_worker.api_client.add_response(
        "CreateImage",
        body={"hash": "01e9327d0e5ab7c516a0792c3662bc65"},
        response={
            "id": "bbbbbbbb-aaaa-3333-aaaa-aaaaaaaaaaaa",
            "s3_put_url": "http://s3/put",
            "s3_url": "http://s3/image1.jpg",
        },
    )
    mock_worker.api_client.add_response(
        "CreateImage",
        body={"hash": "01e9327d0e5ab7c516a0792c3662bc65"},
        response={
            "id": "bbbbbbbb-aaaa-4444-aaaa-aaaaaaaaaaaa",
            "s3_put_url": "http://s3/put",
            "s3_url": "http://s3/image2.jpg",
        },
    )
    mock_worker.api_client.add_response(
        "CreateImage",
        body={"hash": "01e9327d0e5ab7c516a0792c3662bc65"},
        response={
            "id": "bbbbbbbb-aaaa-5555-aaaa-aaaaaaaaaaaa",
            "s3_put_url": "http://s3/put",
            "s3_url": "http://s3/image3.jpg",
        },
    )
    mock_worker.api_client.add_response(
        "CreateImage",
        body={"hash": "01e9327d0e5ab7c516a0792c3662bc65"},
        response={
            "id": "bbbbbbbb-aaaa-6666-aaaa-aaaaaaaaaaaa",
            "s3_put_url": "http://s3/put",
            "s3_url": "http://s3/image4.jpg",
        },
    )
    responses.add(responses.PUT, "http://s3/put", status=201)
    responses.add(responses.PUT, "http://s3/put", status=201)
    responses.add(responses.PUT, "http://s3/put", status=201)
    responses.add(responses.PUT, "http://s3/put", status=201)
    mock_worker.api_client.add_response(
        "PartialUpdateImage",
        id="bbbbbbbb-aaaa-4444-aaaa-aaaaaaaaaaaa",
        body={"status": "checked"},
        response={"id": "bbbbbbbb-aaaa-4444-aaaa-aaaaaaaaaaaa", "status": "checked"},
    )
    mock_worker.api_client.add_response(
        "PartialUpdateImage",
        id="bbbbbbbb-aaaa-3333-aaaa-aaaaaaaaaaaa",
        body={"status": "checked"},
        response={"id": "bbbbbbbb-aaaa-3333-aaaa-aaaaaaaaaaaa", "status": "checked"},
    )
    mock_worker.api_client.add_response(
        "PartialUpdateImage",
        id="bbbbbbbb-aaaa-5555-aaaa-aaaaaaaaaaaa",
        body={"status": "checked"},
        response={"id": "bbbbbbbb-aaaa-5555-aaaa-aaaaaaaaaaaa", "status": "checked"},
    )
    mock_worker.api_client.add_response(
        "PartialUpdateImage",
        id="bbbbbbbb-aaaa-6666-aaaa-aaaaaaaaaaaa",
        body={"status": "checked"},
        response={"id": "bbbbbbbb-aaaa-6666-aaaa-aaaaaaaaaaaa", "status": "checked"},
    )

    # Page creation
    mock_worker.api_client.add_response(
        "CreateElement",
        body={
            "parent": "folderid",
            "corpus": "corpusid",
            "type": "book",
            "name": "file.pdf",
            "worker_run_id": mock_worker.worker_run_id,
        },
        response={
            "id": "box1",
            "type": "book",
            "corpus": {"id": "corpusid"},
        },
    )
    mock_worker.api_client.add_response(
        "CreateElement",
        body={
            "parent": "folderid",
            "corpus": "corpusid",
            "type": "book",
            "name": "file2.pdf",
            "worker_run_id": mock_worker.worker_run_id,
        },
        response={
            "id": "box2",
            "type": "book",
            "corpus": {"id": "corpusid"},
        },
    )
    mock_worker.api_client.add_response(
        "CreateElement",
        body={
            "type": "page",
            "name": "1",
            "image": "bbbbbbbb-aaaa-5555-aaaa-aaaaaaaaaaaa",
            "corpus": "corpusid",
            "polygon": None,
            "parent": "box1",
            "worker_run_id": "bbbbbbbb-aaaa-aaaa-aaaa-aaaaaaaaaaaa",
            "confidence": None,
        },
        response={
            "id": "page1",
            "type": "page",
            "zone": {
                "image": {
                    "width": 1200,
                    "height": 3000,
                }
            },
        },
    )
    mock_worker.api_client.add_response(
        "CreateElement",
        body={
            "type": "page",
            "name": "2",
            "image": "bbbbbbbb-aaaa-6666-aaaa-aaaaaaaaaaaa",
            "corpus": "corpusid",
            "polygon": None,
            "parent": "box1",
            "worker_run_id": "bbbbbbbb-aaaa-aaaa-aaaa-aaaaaaaaaaaa",
            "confidence": None,
        },
        response={
            "id": "page2",
            "type": "page",
            "zone": {
                "image": {
                    "width": 1200,
                    "height": 3000,
                }
            },
        },
    )
    mock_worker.api_client.add_response(
        "CreateElement",
        body={
            "type": "page",
            "name": "1",
            "image": "bbbbbbbb-aaaa-3333-aaaa-aaaaaaaaaaaa",
            "corpus": "corpusid",
            "polygon": None,
            "parent": "box2",
            "worker_run_id": mock_worker.worker_run_id,
            "confidence": None,
        },
        response={
            "id": "page3",
            "type": "page",
            "zone": {
                "image": {
                    "width": 1200,
                    "height": 3000,
                }
            },
        },
    )
    mock_worker.api_client.add_response(
        "CreateElement",
        body={
            "type": "page",
            "name": "2",
            "image": "bbbbbbbb-aaaa-4444-aaaa-aaaaaaaaaaaa",
            "corpus": "corpusid",
            "polygon": None,
            "parent": "box2",
            "worker_run_id": mock_worker.worker_run_id,
            "confidence": None,
        },
        response={
            "id": "page4",
            "type": "page",
            "zone": {
                "image": {
                    "width": 1200,
                    "height": 3000,
                }
            },
        },
    )

    # Process update
    mock_worker.api_client.add_response(
        "PartialUpdateProcess",
        id="processid",
        body={"element_id": "folderid"},
        response={},
    )

    # DataFile deletion
    mock_worker.api_client.add_response("DestroyDataFile", id="file1", response={})
    mock_worker.api_client.add_response("DestroyDataFile", id="file2", response={})

    # Transcriptions
    mock_worker.api_client.add_response(
        "CreateElementTranscriptions",
        id="page1",
        body={
            "element_type": "text_line",
            "worker_run_id": mock_worker.worker_run_id,
            "transcriptions": [
                {
                    "text": " A Simple PDF File \n",
                    "polygon": [[112, 285], [574, 285], [574, 183], [112, 183]],
                    "confidence": 1,
                    "orientation": "horizontal-lr",
                },
                {
                    "text": " This is a small demonstration .pdf file - \n",
                    "polygon": [[136, 399], [481, 399], [481, 362], [136, 362]],
                    "confidence": 1,
                    "orientation": "horizontal-lr",
                },
                {
                    "text": " just for use in the Virtual Mechanics tutorials. More text. And more \n",
                    "polygon": [[136, 490], [719, 490], [719, 452], [136, 452]],
                    "confidence": 1,
                    "orientation": "horizontal-lr",
                },
                {
                    "text": " text. And more text. And more text. And more text. \n",
                    "polygon": [[136, 535], [583, 535], [583, 497], [136, 497]],
                    "confidence": 1,
                    "orientation": "horizontal-lr",
                },
                {
                    "text": " And more text. And more text. And more text. And more text. And more \n",
                    "polygon": [[136, 626], [764, 626], [764, 588], [136, 588]],
                    "confidence": 1,
                    "orientation": "horizontal-lr",
                },
                {
                    "text": " text. And more text. Boring, zzzzz. And more text. And more text. And \n",
                    "polygon": [[136, 671], [750, 671], [750, 633], [136, 633]],
                    "confidence": 1,
                    "orientation": "horizontal-lr",
                },
                {
                    "text": " more text. And more text. And more text. And more text. And more text. \n",
                    "polygon": [[136, 716], [766, 716], [766, 679], [136, 679]],
                    "confidence": 1,
                    "orientation": "horizontal-lr",
                },
                {
                    "text": " And more text. And more text. \n",
                    "polygon": [[136, 762], [407, 762], [407, 724], [136, 724]],
                    "confidence": 1,
                    "orientation": "horizontal-lr",
                },
                {
                    "text": " And more text. And more text. And more text. And more text. And more \n",
                    "polygon": [[136, 852], [764, 852], [764, 814], [136, 814]],
                    "confidence": 1,
                    "orientation": "horizontal-lr",
                },
                {
                    "text": " text. And more text. And more text. Even more. Continued on page 2 ...\n",
                    "polygon": [[136, 897], [759, 897], [759, 860], [136, 860]],
                    "confidence": 1,
                    "orientation": "horizontal-lr",
                },
            ],
            "return_elements": True,
        },
        response={},
    )
    mock_worker.api_client.add_response(
        "CreateElementTranscriptions",
        id="page2",
        body={
            "element_type": "text_line",
            "worker_run_id": mock_worker.worker_run_id,
            "transcriptions": [
                {
                    "text": " Simple PDF File 2 \n",
                    "polygon": [[112, 285], [568, 285], [568, 183], [112, 183]],
                    "confidence": 1,
                    "orientation": "horizontal-lr",
                },
                {
                    "text": " ...continued from page 1. Yet more text. And more text. And more text. \n",
                    "polygon": [[136, 399], [757, 399], [757, 362], [136, 362]],
                    "confidence": 1,
                    "orientation": "horizontal-lr",
                },
                {
                    "text": " And more text. And more text. And more text. And more text. And more \n",
                    "polygon": [[136, 445], [764, 445], [764, 407], [136, 407]],
                    "confidence": 1,
                    "orientation": "horizontal-lr",
                },
                {
                    "text": " text. Oh, how boring typing this stuff. But not as boring as watching \n",
                    "polygon": [[136, 490], [727, 490], [727, 452], [136, 452]],
                    "confidence": 1,
                    "orientation": "horizontal-lr",
                },
                {
                    "text": " paint dry. And more text. And more text. And more text. And more text. \n",
                    "polygon": [[136, 535], [759, 535], [759, 497], [136, 497]],
                    "confidence": 1,
                    "orientation": "horizontal-lr",
                },
                {
                    "text": " Boring.  More, a little more text. The end, and just as well. \n",
                    "polygon": [[136, 581], [647, 581], [647, 543], [136, 543]],
                    "confidence": 1,
                    "orientation": "horizontal-lr",
                },
            ],
            "return_elements": True,
        },
        response={},
    )
    mock_worker.api_client.add_response(
        "CreateElementTranscriptions",
        id="page3",
        body={
            "element_type": "text_line",
            "worker_run_id": mock_worker.worker_run_id,
            "return_elements": True,
            "transcriptions": [
                {
                    "text": "Mon cher Marc,\n",
                    "polygon": [[145, 302], [299, 302], [299, 262], [145, 262]],
                    "confidence": 1,
                    "orientation": "horizontal-lr",
                },
                {
                    "text": "Je suis descendu ce matin chez mon médecin Hermogène, qui vient de rentrer à la Villa\n",
                    "polygon": [[145, 353], [1006, 353], [1006, 314], [145, 314]],
                    "confidence": 1,
                    "orientation": "horizontal-lr",
                },
                {
                    "text": "après un assez long voyage en Asie. L’examen devait se faire à jeun : nous avions pris\n",
                    "polygon": [[145, 405], [995, 405], [995, 366], [145, 366]],
                    "confidence": 1,
                    "orientation": "horizontal-lr",
                },
                {
                    "text": "rendez-vous pour les premières heures de la matinée. Je me suis couché sur un lit après\n",
                    "polygon": [[145, 457], [1016, 457], [1016, 418], [145, 418]],
                    "confidence": 1,
                    "orientation": "horizontal-lr",
                },
                {
                    "text": "m’être dépouillé de mon manteau et de ma tunique. Je t’épargne des détails qui te seraient\n",
                    "polygon": [[145, 509], [1036, 509], [1036, 470], [145, 470]],
                    "confidence": 1,
                    "orientation": "horizontal-lr",
                },
                {
                    "text": "aussi désagréables qu’à moi-même, et la description du corps d’un homme qui avance en\n",
                    "polygon": [[145, 561], [1025, 561], [1025, 522], [145, 522]],
                    "confidence": 1,
                    "orientation": "horizontal-lr",
                },
                {
                    "text": "âge et s’apprête à mourir d’une hydropisie du cœur. Disons seulement que j’ai toussé,\n",
                    "polygon": [[145, 613], [988, 613], [988, 573], [145, 573]],
                    "confidence": 1,
                    "orientation": "horizontal-lr",
                },
                {
                    "text": "respiré, et retenu mon souffle selon les indications d’Hermogène, alarmé malgré lui par les\n",
                    "polygon": [[145, 664], [1030, 664], [1030, 625], [145, 625]],
                    "confidence": 1,
                    "orientation": "horizontal-lr",
                },
                {
                    "text": "progrès si rapides du mal, et prêt à en rejeter le blâme sur le jeune Iollas qui m’a soigné en\n",
                    "polygon": [[145, 716], [1034, 716], [1034, 677], [145, 677]],
                    "confidence": 1,
                    "orientation": "horizontal-lr",
                },
                {
                    "text": "son absence. Il est difficile de rester empereur en présence d’un médecin, et difficile aussi\n",
                    "polygon": [[145, 768], [1025, 768], [1025, 729], [145, 729]],
                    "confidence": 1,
                    "orientation": "horizontal-lr",
                },
                {
                    "text": "de garder sa qualité d’homme\n",
                    "polygon": [[145, 820], [438, 820], [438, 781], [145, 781]],
                    "confidence": 1,
                    "orientation": "horizontal-lr",
                },
            ],
        },
        response={},
    )
    mock_worker.api_client.add_response(
        "CreateElementTranscriptions",
        id="page4",
        body={
            "element_type": "text_line",
            "worker_run_id": mock_worker.worker_run_id,
            "transcriptions": [
                {
                    "text": "Animula vagula blandula\n",
                    "polygon": [[205, 302], [447, 302], [447, 262], [205, 262]],
                    "confidence": 1,
                    "orientation": "horizontal-lr",
                },
                {
                    "text": "Hospes comesque corporis\n",
                    "polygon": [[205, 396], [474, 396], [474, 357], [205, 357]],
                    "confidence": 1,
                    "orientation": "horizontal-lr",
                },
                {
                    "text": "Quae nunc abibis in loca\n",
                    "polygon": [[205, 491], [448, 491], [448, 452], [205, 452]],
                    "confidence": 1,
                    "orientation": "horizontal-lr",
                },
                {
                    "text": "Pallidula rigida nudula\n",
                    "polygon": [[205, 585], [422, 585], [422, 546], [205, 546]],
                    "confidence": 1,
                    "orientation": "horizontal-lr",
                },
                {
                    "text": "Nec ut soles dabis iocos\n",
                    "polygon": [[205, 680], [444, 680], [444, 641], [205, 641]],
                    "confidence": 1,
                    "orientation": "horizontal-lr",
                },
            ],
            "return_elements": True,
        },
        response={},
    )

    mock_worker.run()

    with (mock_worker.work_dir / "elements.json").open() as f:
        elements = json.load(f)
    assert elements == [
        {"id": "folderid", "type": "book"},
        {"id": "box2", "type": "book", "corpus": {"id": "corpusid"}},
        {
            "id": "page3",
            "type": "page",
            "zone": {
                "image": {
                    "width": 1200,
                    "height": 3000,
                }
            },
        },
        {
            "id": "page4",
            "type": "page",
            "zone": {
                "image": {
                    "width": 1200,
                    "height": 3000,
                }
            },
        },
        {"id": "box1", "type": "book", "corpus": {"id": "corpusid"}},
        {
            "id": "page1",
            "type": "page",
            "zone": {
                "image": {
                    "width": 1200,
                    "height": 3000,
                }
            },
        },
        {
            "id": "page2",
            "type": "page",
            "zone": {
                "image": {
                    "width": 1200,
                    "height": 3000,
                }
            },
        },
    ]
